@@ -30,10 +30,11 @@
 #include "mdnsd.h"
 #include <string.h>
 #include <stdlib.h>
+#include <stdbool.h>
 #include <time.h>
 #include <errno.h>
 
-#define SPRIME 108		/* Size of query/publish hashes */
+#define SPRIME 109		/* Size of query/publish hashes */
 #define LPRIME 1009		/* Size of cache hash */
 
 #define GC 86400                /* Brute force garbage cleanup
@@ -98,6 +99,7 @@ struct mdns_daemon {
 	struct query *queries[SPRIME], *qlist;
 
 	struct in_addr addr;
+	struct in6_addr addr_v6;
 
 	mdnsd_record_received_callback received_callback;
 	void *received_callback_data;
@@ -174,6 +176,8 @@ static size_t _rr_len(mdns_answer_t *rr)
 		len += strlen(rr->rdname); /* worst case */
 	if (rr->ip.s_addr)
 		len += 4;
+	if (! IN6_IS_ADDR_UNSPECIFIED(&(rr->ip6)))
+		len += 16;
 	if (rr->type == QTYPE_PTR)
 		len += 6;	/* srv record stuff */
 
@@ -181,29 +185,35 @@ static size_t _rr_len(mdns_answer_t *rr)
 }
 
 /* Compares new rdata with known a, painfully */
-static int _a_match(struct resource *r, mdns_answer_t *a)
+static bool _a_match(struct resource *r, mdns_answer_t *a)
 {
 	if (!a->name)
 		return 0;
 	if (strcmp(r->name, a->name) || r->type != a->type)
 		return 0;
 
-	if (r->type == QTYPE_SRV &&
-	    r->known.srv.name && a->rdname && !strcmp(r->known.srv.name, a->rdname) &&
-	    a->srv.port == r->known.srv.port &&
-	    a->srv.weight == r->known.srv.weight &&
-	    a->srv.priority == r->known.srv.priority)
-		return 1;
+	switch (r->type) {
+		case QTYPE_SRV:
+			return r->known.srv.name && a->rdname && strcmp(r->known.srv.name, a->rdname) == 0
+				&& a->srv.port == r->known.srv.port
+				&& a->srv.weight == r->known.srv.weight
+				&& a->srv.priority == r->known.srv.priority;
 
-	if ((r->type == QTYPE_PTR || r->type == QTYPE_NS || r->type == QTYPE_CNAME) &&
-	    !strcmp(a->rdname, r->known.ns.name))
-		return 1;
 
-	if (r->type == QTYPE_A || !memcmp(&r->known.a.ip, &a->ip, 4))
-		return 1;
+		case QTYPE_PTR:
+		case QTYPE_NS:
+		case QTYPE_CNAME:
+			return strcmp(a->rdname, r->known.ns.name) == 0;
 
-	if (r->rdlength == a->rdlen && !memcmp(r->rdata, a->rdata, r->rdlength))
-		return 1;
+		case QTYPE_A:
+			return memcmp(&r->known.a.ip, &a->ip, 4) == 0;
+
+		case QTYPE_AAAA:
+			return memcmp(&r->known.aaaa.ip6, &a->ip6, 16) == 0;
+
+		default:
+			return r->rdlength == a->rdlen && memcmp(r->rdata, a->rdata, r->rdlength) == 0;
+	}
 
 	return 0;
 }
@@ -381,12 +391,18 @@ static void _free_cached(struct cached *c)
 	if (!c)
 		return;
 
-	if (c->rr.name)
+	if (c->rr.name) {
 		free(c->rr.name);
-	if (c->rr.rdata)
+		c->rr.name = NULL;
+	}
+	if (c->rr.rdata) {
 		free(c->rr.rdata);
-	if (c->rr.rdname)
+		c->rr.rdata = NULL;
+	}
+	if (c->rr.rdname) {
 		free(c->rr.rdname);
+		c->rr.rdname = NULL;
+	}
 	free(c);
 }
 
@@ -395,12 +411,18 @@ static void _free_record(mdns_record_t *r)
 	if (!r)
 		return;
 
-	if (r->rr.name)
-	    free(r->rr.name);
-	if (r->rr.rdata)
+	if (r->rr.name) {
+		free(r->rr.name);
+		r->rr.name = NULL;
+	}
+	if (r->rr.rdata) {
 		free(r->rr.rdata);
-	if (r->rr.rdname)
+		r->rr.rdata = NULL;
+	}
+	if (r->rr.rdname) {
 		free(r->rr.rdname);
+		r->rr.rdname = NULL;
+	}
 	free(r);
 }
 
@@ -561,6 +583,10 @@ static int _cache(mdns_daemon_t *d, struct resource *r, struct in_addr ip)
 		c->rr.ip = r->known.a.ip;
 		break;
 
+	case QTYPE_AAAA:
+		c->rr.ip6 = r->known.aaaa.ip6;
+		break;
+
 	case QTYPE_NS:
 	case QTYPE_CNAME:
 	case QTYPE_PTR:
@@ -594,7 +620,9 @@ static void _a_copy(struct message *m, mdns_answer_t *a)
 	}
 
 	if (a->ip.s_addr)
-		message_rdata_raw(m, (unsigned char *)&a->ip, 4);
+		message_rdata_ipv4(m, a->ip);
+	else if (!IN6_IS_ADDR_UNSPECIFIED(&(a->ip6)))
+		message_rdata_ipv6(m, a->ip6);
 	if (a->type == QTYPE_SRV)
 		message_rdata_srv(m, a->srv.priority, a->srv.weight, a->srv.port, a->rdname);
 	else if (a->rdname)
@@ -679,8 +707,15 @@ void mdnsd_set_address(mdns_daemon_t *d, struct in_addr addr)
 		while (r) {
 			next = r->next;
 
-			if (r->rr.type == QTYPE_A)
-				mdnsd_set_ip(d, r, addr);
+			if (r->rr.type == QTYPE_A) {
+				if (addr.s_addr == 0) {
+					r->rr.ttl = 0;
+					r->list = d->a_now;
+					d->a_now = r;
+				} else {
+					mdnsd_set_ip(d, r, addr);
+				}
+			}
 
 			r = next;
 		}
@@ -694,6 +729,42 @@ struct in_addr mdnsd_get_address(mdns_daemon_t *d)
 	return d->addr;
 }
 
+void mdnsd_set_ipv6_address(mdns_daemon_t *d, struct in6_addr addr)
+{
+	int i;
+
+	if (!memcmp(&d->addr_v6, &addr, sizeof(d->addr_v6)))
+		return;		/* No change */
+
+	for (i = 0; i < SPRIME; i++) {
+		mdns_record_t *r, *next;
+
+		r = d->published[i];
+		while (r) {
+			next = r->next;
+
+			if (r->rr.type == QTYPE_AAAA) {
+				if (IN6_IS_ADDR_UNSPECIFIED(&addr)) {
+					r->rr.ttl = 0;
+					r->list = d->a_now;
+					d->a_now = r;
+				} else {
+					mdnsd_set_ipv6(d, r, addr);
+				}
+			}
+
+			r = next;
+		}
+	}
+
+	d->addr_v6 = addr;
+}
+
+struct in6_addr mdnsd_get_ipv6_address(mdns_daemon_t *d)
+{
+	return d->addr_v6;
+}
+
 /* Shutting down, zero out ttl and push out all records */
 void mdnsd_shutdown(mdns_daemon_t *d)
 {
@@ -740,6 +811,7 @@ void mdnsd_free(mdns_daemon_t *d)
 		while (cur) {
 			struct cached *next = cur->next;
 
+			cur->next = NULL;
 			_free_cached(cur);
 			cur = next;
 		}
@@ -752,6 +824,7 @@ void mdnsd_free(mdns_daemon_t *d)
 		while (cur) {
 			struct mdns_record *next = cur->next;
 
+			cur->next = NULL;
 			_free_record(cur);
 			cur = next;
 		}
@@ -760,6 +833,7 @@ void mdnsd_free(mdns_daemon_t *d)
 		while (curq) {
 			struct query *next = curq->next;
 
+			curq->next = NULL;
 			free(curq->name);
 			free(curq);
 			curq = next;
@@ -770,6 +844,7 @@ void mdnsd_free(mdns_daemon_t *d)
 	while (u) {
 		struct unicast *next = u->next;
 
+		u->next = NULL;
 		free(u);
 		u = next;
 	}
@@ -830,7 +905,7 @@ int mdnsd_in(mdns_daemon_t *d, struct message *m, struct in_addr ip, unsigned sh
 				/* probing state, check for conflicts */
 				if (r->unique && r->unique < 5 && !r->modified) {
 					/* Check all to-be answers against our own */
-					for (j = 0; j < m->nscount; j++) {
+					for (j = 0; j < m->ancount; j++) {
 						if (!m->an || m->qd[i].type != m->an[j].type || strcmp(m->qd[i].name, m->an[j].name))
 							continue;
 
@@ -925,7 +1000,7 @@ int mdnsd_out(mdns_daemon_t *d, struct message *m, struct in_addr *ip, unsigned
 		INFO("Send Unicast Answer: Name: %s, Type: %d", u->r->rr.name, u->r->rr.type);
 
 		d->uanswers = u->next;
-		*port = u->port;
+		*port = htons(u->port);
 		*ip = u->to;
 		m->id = u->id;
 		message_qd(m, u->r->rr.name, u->r->rr.type, d->class);
@@ -1372,6 +1447,12 @@ void mdnsd_set_ip(mdns_daemon_t *d, mdns_record_t *r, struct in_addr ip)
 	_r_publish(d, r);
 }
 
+void mdnsd_set_ipv6(mdns_daemon_t *d, mdns_record_t *r, struct in6_addr ip6)
+{
+	r->rr.ip6 = ip6;
+	_r_publish(d, r);
+}
+
 void mdnsd_set_srv(mdns_daemon_t *d, mdns_record_t *r, unsigned short priority, unsigned short weight, unsigned short port, char *name)
 {
 	r->rr.srv.priority = priority;
@@ -1454,7 +1535,7 @@ int mdnsd_step(mdns_daemon_t *d, int sd, bool in, bool out, struct timeval *tv)
 	}
 
 	/* Service Enumeration/Discovery completed */
-	if (d->disco)
+	if (d && d->disco)
 		d->disco = 0;
 
 	return rc;
